import os
import signal
import subprocess
from subprocess import TimeoutExpired

from omniduct.utils.config import config as omniduct_config
from omniduct.utils.debug import logger


__all__ = ["run_in_subprocess", "TimeoutExpired", "Timeout"]

DEFAULT_SUBPROCESS_CONFIG = {
    "shell": True,
    "close_fds": False,
    "stdin": None,
    "stdout": subprocess.PIPE,
    "stderr": subprocess.PIPE,
    "preexec_fn": os.setsid,  # Set the process as the group leader, so we can kill recursively
}


class SubprocessResults:
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)


def run_in_subprocess(cmd, check_output=False, **kwargs):
    """
    Execute command using default subprocess configuration.

    Parameters
    ----------
    cmd : string
        Command to be executed in subprocess.
    kwargs : keywords
        Options to pass to subprocess.Popen.

    Returns
    -------
    proc : Popen subprocess
        Subprocess used to run command.
    """

    with open(os.devnull, "w", encoding="utf-8") as devnull:
        logger.debug(f"Executing command: {cmd}")
        config = DEFAULT_SUBPROCESS_CONFIG.copy()
        config.update(kwargs)
        if not check_output:
            if omniduct_config.logging_level < 20:
                config["stdout"] = None
                config["stderr"] = None
            else:
                config["stdout"] = devnull
                config["stderr"] = devnull
        timeout = config.pop("timeout", None)

        with subprocess.Popen(cmd, **config) as process:
            try:
                stdout, stderr = process.communicate(None, timeout=timeout)
                returncode = process.returncode
            except subprocess.TimeoutExpired as e:
                os.killpg(
                    os.getpgid(process.pid), signal.SIGINT
                )  # send signal to the process group, recursively killing all children
                output, unused_err = process.communicate()
                raise subprocess.TimeoutExpired(
                    process.args, timeout, output=output
                ) from e
        return SubprocessResults(
            returncode=returncode, stdout=stdout or b"", stderr=stderr or b""
        )


class Timeout:
    def __init__(self, seconds=1, error_message="Timeout"):
        self.seconds = seconds
        self.error_message = error_message

    def handle_timeout(self, signum, frame):
        raise TimeoutError(self.error_message)

    def __enter__(self):
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self.seconds)

    # pylint: disable-next=redefined-builtin
    def __exit__(self, type, value, traceback):
        signal.alarm(0)



import os
import yaml

from omniduct import config, logger, DuctRegistry


SERVICES_CONFIG_PATH = os.path.join(os.path.dirname(__file__), 'services.yml')

# Set up Omniduct configuration path
config._config_path = '~/.example_wrapper/config'

# Build registry from configuration
# Note: If you need to transform the configuration before importing, you can
# directly load the configuration into a dictionary using the `yaml` package,
# modify it before passing it in as the configuration below.
registry = DuctRegistry(config=SERVICES_CONFIG_PATH)



import random
import re
import socket

from omniduct.utils.debug import logger


def is_local_port_free(local_port):
    """
    Checks if local port is free.

    Parameters
    ----------
    local_port : int
        Local port to check.

    Returns
    -------
    out : boolean
        Whether local port is free.
    """
    s = socket.socket()
    try:
        s.bind(("", local_port))
    except socket.error:
        return False
    finally:
        s.close()
    return True


def get_free_local_port():
    """
    Return a random free port

    Returns
    -------
    free_port : int
        A free local port
    """
    s = socket.socket()
    s.bind(("", 0))
    free_port = s.getsockname()[1]
    s.close()
    logger.info(f"found port {free_port}")
    return free_port


def is_port_bound(hostname, port, timeout=None):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    if timeout:
        s.settimeout(timeout)
    try:
        s.connect((hostname, port))
    except:  # pylint: disable=bare-except
        return False
    finally:
        s.close()
    return True


# Random hosts for ssh gateway nodes
def naive_load_balancer(hosts, port):
    # Shuffle hosts randomly
    hosts = hosts.copy()
    random.shuffle(hosts)

    # Check if host is available and if so return it
    pattern = re.compile(r"(?P<host>[^\:]+)(?::(?P<port>[0-9]{1,5}))?")
    for host in hosts:
        m = pattern.match(host)
        if is_port_bound(m.group("host"), int(m.group("port") or port), timeout=1):
            return host
        logger.warning(f"Avoiding down or inaccessible host: '{host}'.")

    raise RuntimeError(
        "Unable to connect to any of the hosts associated with this service. "
        "This may be due to networking issues, such as not being connected to "
        "the internet or your company's VPN."
    )



import inspect

import decorator


def function_args_as_kwargs(func, *args, **kwargs):
    arguments = inspect.signature(func).parameters.keys()
    kwargs.update(dict(zip(list(arguments), args)))
    return kwargs


@decorator.decorator
def require_connection(f, self, *args, **kwargs):
    """
    A wrapper to allow restoring of connection status in the event that
    connection issues result in failures. If so, we will attempt to retry the
    failed function call once more.
    """
    if not self._Duct__connected:
        self.connect()

    try:
        return f(self, *args, **kwargs)
    except Exception:  # pylint: disable=broad-exception-caught
        # Check to see if it is possible that we failed due to connection issues.
        # If so, try again once more. If we fail again, raise.
        # TODO: Explore adding a DuctConnectionError class and filter this
        # handling to errors of that class.
        if not self.is_connected():
            self.connect()
            return f(self, *args, **kwargs)
        raise



class TreeProxy:
    """
    A read-only proxy object for a dictionary tree structure that allows accessing
    of keys via attributes and indexing.

    Once parsed using the `_for_dict` constructor, or if used directly using the
    `_for_tree` constructor, trees should have form:

    ```
    {'key': {'nested': {None: <object>}}}
    ```
    """

    __slots__ = ("__tree__", "__nodename__")

    @classmethod
    def _for_dict(cls, dct, key_parser=None, name=None):
        return cls._for_tree(cls.__dict_to_tree(dct, key_parser=key_parser), name=name)

    @classmethod
    def _for_tree(cls, tree, name=None):
        if None in tree:
            return tree[None]
        return cls(tree, name=name)

    def __init__(self, tree, name=None):
        self.__tree__ = tree
        self.__nodename__ = str(name) if name else None

    def __getitem__(self, name):
        if name in self.__tree__:
            if not isinstance(self.__tree__[name], TreeProxy):
                return TreeProxy._for_tree(
                    self.__tree__[name], name=self.__name_of_child(name)
                )
            return self.__tree__[name]
        raise KeyError(f"Invalid child node `{name}`.")

    def __iter__(self):
        return iter(self.__tree__)

    def __len__(self):
        return len(self.__tree__)

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError as e:
            raise AttributeError(f"Invalid child node `{name}`.") from e

    def __dir__(self):
        return list(self.__tree__)

    def __repr__(self):
        if self.__nodename__:
            return (
                f"<TreeProxy of '{self.__nodename__}' with {len(self.__tree__)} nodes>"
            )
        return f"<TreeProxy of dictionary with {len(self.__tree__)} nodes>"

    # Helpers

    def __name_of_child(self, child):
        if self.__nodename__:
            return ".".join([self.__nodename__, str(child)])
        return str(child)

    @classmethod
    def __dict_to_tree(cls, dct, key_parser):
        tree = {}
        for key, value in dct.items():
            cls.__add_nested_key_value(
                tree, keys=key_parser(key, value) if key_parser else [key], value=value
            )
        return tree

    @classmethod
    def __add_nested_key_value(cls, tree, keys, value):
        for key in keys:
            if key not in tree:
                tree[key] = {}
            tree = tree[key]
        if len(tree) and None not in tree:
            raise ValueError(
                f"`TreeProxy` objects can only proxy trees with values only on leaf nodes; error encounted while trying to add value to node {keys}."
            )
        tree[None] = value



# pylint: disable=abstract-method

import posixpath
import select
import socketserver
import stat
import threading

from interface_meta import override

from omniduct.errors import DuctAuthenticationError
from omniduct.filesystems.base import FileSystemFileDesc
from omniduct.remotes.base import RemoteClient
from omniduct.utils.debug import logger
from omniduct.utils.processes import SubprocessResults

__all__ = ["ParamikoSSHClient"]


class ParamikoSSHClient(RemoteClient):
    """
    An experimental SSH client that uses a `paramiko` rather than command-line
    SSH backend. This client has been fully implemented and should work as is,
    but until it receives further testing, we recommend using the cli backed SSH
    client.
    """

    PROTOCOLS = ["ssh_paramiko"]
    DEFAULT_PORT = 22

    @override
    def _init(self):
        logger.warning(
            "The Paramiko SSH client is still under development, \
                        and is not ready for use as a daily driver."
        )

    @override
    def _connect(self):
        import paramiko

        # pylint: disable-next=attribute-defined-outside-init
        self.__client = paramiko.SSHClient()
        self.__client.set_missing_host_key_policy(paramiko.client.AutoAddPolicy())
        self.__client.load_system_host_keys()

        try:
            self.__client.connect(self.host, username=self.username)
            # pylint: disable-next=attribute-defined-outside-init
            self.__client_sftp = paramiko.SFTPClient.from_transport(
                self.__client.get_transport()
            )
        except paramiko.SSHException as e:
            if len(e.args) == 1 and e.args[0] == "No authentication methods available":
                raise DuctAuthenticationError(e.args[0]) from e
            raise e

    @override
    def _is_connected(self):
        try:
            return self.__client.get_transport().is_active()
        except:  # pylint: disable=bare-except
            return False

    @override
    def _disconnect(self):
        try:
            self.__client_sftp.close()
            self.__client.close()
        except:  # pylint: disable=bare-except
            pass

    @override
    def _execute(self, cmd, **kwargs):
        _, stdout, stderr = self.__client.exec_command(cmd)
        returncode = stdout.channel.recv_exit_status()
        return SubprocessResults(
            returncode=returncode, stdout=stdout.read(), stderr=stderr.read()
        )

    @override
    def _port_forward_start(self, local_port, remote_host, remote_port):
        logger.debug(
            f"Now forwarding port {local_port} to {remote_host}:{remote_port} ..."
        )

        try:
            server = forward_tunnel(
                local_port, remote_host, remote_port, self.__client.get_transport()
            )
        except KeyboardInterrupt:
            print("C-c: Port forwarding stopped.")
        return server

    @override
    def _port_forward_stop(self, local_port, remote_host, remote_port, connection):
        connection.shutdown()

    @override
    def _is_port_bound(self, host, port):
        return True

    # Path properties and helpers
    @override
    def _path_home(self):
        return self.execute("echo ~", skip_cwd=True).stdout.decode("utf-8").strip()

    @override
    def _path_separator(self):
        return "/"

    # File node properties
    @override
    def _exists(self, path):
        try:
            self.__client_sftp.stat(path)
            return True
        except FileNotFoundError:
            return False

    @override
    def _isdir(self, path):
        try:
            return stat.S_ISDIR(self.__client_sftp.stat(path).st_mode)
        except FileNotFoundError:
            return False

    @override
    def _isfile(self, path):
        try:
            return not stat.S_ISDIR(self.__client_sftp.stat(path).st_mode)
        except FileNotFoundError:
            return False

    # Directory handling and enumeration
    @override
    def _dir(self, path):
        for attrs in self.__client_sftp.listdir_attr(path):
            yield FileSystemFileDesc(
                fs=self,
                path=posixpath.join(path, attrs.filename),
                name=attrs.filename,
                type="directory"
                if stat.S_ISDIR(attrs.st_mode)
                else "file",  # TODO: What about links, which are of form: lrwxrwxrwx?
                bytes=attrs.st_size,
                owner=attrs.st_uid,
                group=attrs.st_gid,
                last_modified=attrs.st_mtime,
            )

    @override
    def _mkdir(self, path, recursive, exist_ok):
        if exist_ok and self.isdir(path):
            return
        assert (
            self.execute(
                "mkdir " + ("-p " if recursive else "") + f'"{path}"'
            ).returncode
            == 0
        ), f"Failed to create directory at: `{path}`"

    @override
    def _remove(self, path, recursive):
        assert (
            self.execute(
                "rm -f " + ("-r " if recursive else "") + f'"{path}"'
            ).returncode
            == 0
        ), f"Failed to remove file(s) at: `{path}`"

    # File handling
    @override
    def _open(self, path, mode):
        """
        Paramiko offers a complete file-like abstraction for files opened over
        sftp, so we use that abstraction rather than a `FileSystemFile`. Results
        should be indistinguishable.
        """
        return self.__client_sftp.open(path, mode=mode)


# Port Forwarding Utility Code
# Largely based on code from: https://github.com/paramiko/paramiko/blob/master/demos/forward.py


class ForwardServer(socketserver.ThreadingTCPServer):
    daemon_threads = True
    allow_reuse_address = True


class Handler(socketserver.BaseRequestHandler):
    def handle(self):
        try:
            chan = self.ssh_transport.open_channel(
                "direct-tcpip",
                (self.chain_host, self.chain_port),
                self.request.getpeername(),
            )
        except Exception as e:  # pylint: disable=broad-exception-caught
            logger.info(
                "Incoming request to %s:%d failed: %s",
                self.chain_host,
                self.chain_port,
                repr(e),
            )
            return
        if chan is None:
            logger.info(
                "Incoming request to %s:%d was rejected by the SSH server.",
                self.chain_host,
                self.chain_port,
            )
            return

        logger.info(
            f"Connected!  Tunnel open {self.request.getpeername()!r} -> {chan.getpeername()!r} -> {self.chain_host, self.chain_port!r}"
        )
        while True:
            r, _, _ = select.select([self.request, chan], [], [])
            if self.request in r:
                data = self.request.recv(1024)
                if len(data) == 0:
                    break
                chan.send(data)
            if chan in r:
                data = chan.recv(1024)
                if len(data) == 0:
                    break
                self.request.send(data)

        peername = self.request.getpeername()
        chan.close()
        self.request.close()
        logger.info(f"Tunnel closed from {peername!r}")


def forward_tunnel(local_port, remote_host, remote_port, transport):
    # this is a little convoluted, but lets me configure things for the Handler
    # object.  (socketserver doesn't give Handlers any way to access the outer
    # server normally.)
    class SubHandler(Handler):
        chain_host = remote_host
        chain_port = remote_port
        ssh_transport = transport

    server = ForwardServer(("", local_port), SubHandler)

    t = threading.Thread(target=server.serve_forever)
    t.daemon = True  # don't hang on exit
    t.start()

    return server



import unittest
import mock

from omniduct.restful.base import RestClient


class TestRestClient(unittest.TestCase):
    @mock.patch.object(RestClient, "connect")
    @mock.patch("requests.request")
    def test_default_request(self, mock_request, mock_connect):
        client = RestClient(server_protocol="http", host="localhost", port=80)
        client.request("/")
        mock_connect.assert_called_with()
        mock_request.assert_called_with("get", "http://localhost:80/", timeout=None)



# pylint: disable=consider-using-f-string

from __future__ import absolute_import

import ast
import logging
import re

import pandas.io.sql
from interface_meta import override

from omniduct.utils.debug import logger

from .base import DatabaseClient
from ._schemas import SchemasMixin
from . import _pandas


class PrestoClient(DatabaseClient, SchemasMixin):
    """
    This Duct connects to a Facebook Presto server instance using the `pyhive`
    library.

    In addition to the standard `DatabaseClient` API, `PrestoClient` adds a
    `.schemas` descriptor attribute, which enables a tab completion driven
    exploration of a Presto database's schemas and tables.

    Attributes:
        catalog (str): The default catalog to use in database queries.
        schema (str): The default schema/database to use in database queries.
        connection_options (dict): Additional options to pass on to
            `pyhive.presto.connect(...)`.
    """

    PROTOCOLS = ["presto"]
    DEFAULT_PORT = 3506
    SUPPORTS_SESSION_PROPERTIES = True
    NAMESPACE_NAMES = ["catalog", "schema", "table"]
    NAMESPACE_QUOTECHAR = '"'
    NAMESPACE_SEPARATOR = "."

    @property
    @override
    def NAMESPACE_DEFAULT(self):
        return {"catalog": self.catalog, "schema": self.schema}

    @property
    @override
    def NAMESPACE_DEFAULTS_WRITE(self):
        defaults = self.NAMESPACE_DEFAULTS_READ.copy()
        defaults["schema"] = self.username
        return defaults

    @override
    def _init(
        self,
        catalog="default",
        schema="default",
        server_protocol="http",
        source=None,
        requests_session=None,
    ):
        """
        catalog (str): The default catalog to use in database queries.
        schema (str): The default schema/database to use in database queries.
        server_protocol (str): The protocol over which to connect to the Presto REST
            service ('http' or 'https'). (default='http')
        source (str): The source of this query (by default "omniduct <version>").
            If manually specified, result will be: "<source> / omniduct <version>".
        requests_session (requests.Session): an optional requests.Session object for advanced usage.
            Passed through to the pyhive Cursor which supports custom requests sessions for advanced usage
            such as custom headers, cookie values, retry logic, etc.
        """
        self.catalog = catalog
        self.schema = schema
        self.server_protocol = server_protocol
        self.source = source
        self.connection_fields += ("catalog", "schema")
        self._requests_session = requests_session

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, source):
        self._source = source or "omniduct"

    # Connection

    @override
    def _connect(self):
        from sqlalchemy import create_engine

        logging.getLogger("pyhive").setLevel(1000)  # Silence pyhive logging.
        logger.info("Connecting to Presto coordinator...")
        self._sqlalchemy_engine = create_engine(
            f"presto://{self.host}:{self.port}/{self.catalog}/{self.schema}"
        )

    @override
    def _is_connected(self):
        return True

    @override
    def _disconnect(self):
        logger.info("Disconnecting from Presto coordinator...")
        self._sqlalchemy_engine = None
        self._schemas = None  # pylint: disable=attribute-defined-outside-init

    # Querying
    @override
    def _execute(self, statement, cursor, wait, session_properties):
        """
        If something goes wrong, `PrestoClient` will attempt to parse the error
        log and present the user with useful debugging information. If that fails,
        the full traceback will be raised instead.
        """
        from pyhive import presto
        from pyhive.exc import DatabaseError

        try:
            cursor = cursor or presto.Cursor(
                host=self.host,
                port=self.port,
                username=self.username,
                password=self.password,
                catalog=self.catalog,
                schema=self.schema,
                session_props=session_properties,
                poll_interval=1,
                source=self.source,
                protocol=self.server_protocol,
                requests_session=self._requests_session,
            )
            cursor.execute(statement)
            status = cursor.poll()
            if wait:
                logger.progress(0)
                # status None means command executed successfully
                # See https://github.com/dropbox/PyHive/blob/master/pyhive/presto.py#L234
                while status is not None and status["stats"]["state"] != "FINISHED":
                    if status["stats"].get("totalSplits", 0) > 0:
                        pct_complete = round(
                            status["stats"]["completedSplits"]
                            / float(status["stats"]["totalSplits"]),
                            4,
                        )
                        logger.progress(pct_complete * 100)
                    status = cursor.poll()
                logger.progress(100, complete=True)
            return cursor
        except (DatabaseError, pandas.io.sql.DatabaseError) as e:
            try:
                message = e.args[0]
                if isinstance(message, str):
                    message = ast.literal_eval(
                        re.match("[^{]*({.*})[^}]*$", message).group(1)
                    )

                linenumber = message["errorLocation"]["lineNumber"] - 1
                splt = statement.splitlines()
                splt[
                    linenumber
                ] += "   <--  {errorType} ({errorName}) occurred. {message} ".format(
                    **message
                )
                context = "\n\n[Error Context]\n{}\n".format(
                    "\n".join(
                        [
                            splt[ln]
                            for ln in range(
                                max(linenumber - 1, 0), min(linenumber + 2, len(splt))
                            )
                        ]
                    )
                )

                logger.error(context)
            except:  # pylint: disable=bare-except
                logger.warn(
                    (
                        "Omniduct was unable to parse the database error messages. Refer to the "
                        "traceback below for full error details."
                    )
                )

            raise

    @override
    def _query_to_table(self, statement, table, if_exists, **kwargs):
        from pyhive.exc import DatabaseError

        statements = []

        if if_exists == "replace":
            statements.append(f"DROP TABLE IF EXISTS {table};\n")
        elif if_exists == "append":
            raise NotImplementedError(
                f"Append operations have not been implemented for {self.__class__.__name__}."
            )

        statements.append(f"CREATE TABLE {table} AS ({statement})")

        try:
            return self.execute("\n".join(statements), **kwargs)
        except DatabaseError as e:
            if (
                isinstance(e.args, dict)
                and e.args.get("errorName") == "TABLE_ALREADY_EXISTS"
            ):
                raise RuntimeError(f"Table {table} already exists!") from e
            raise

    @override
    def _dataframe_to_table(self, df, table, if_exists="fail", **kwargs):
        """
        If if the schema namespace is not specified, `table.schema` will be
        defaulted to your username. Catalog overrides will be ignored, and will
        default to `self.catalog`.
        """
        return _pandas.to_sql(
            df=df,
            name=table.table,
            schema=table.schema,
            con=self._sqlalchemy_engine,
            index=False,
            if_exists=if_exists,
            **kwargs,
        )

    @override
    def _cursor_empty(self, cursor):
        return False

    @override
    def _table_list(self, namespace, like=None, **kwargs):
        cmd = "SHOW TABLES "
        if namespace:
            cmd = cmd + " FROM " + namespace.name
        if like is not None:
            cmd = cmd + " LIKE " + like + "'"
        return self.query(cmd, **kwargs)

    @override
    def _table_exists(self, table, **kwargs):
        from pyhive.exc import DatabaseError

        logger.disabled = True
        try:
            self.table_desc(table, **kwargs)
            return True
        except DatabaseError:
            return False
        finally:
            logger.disabled = False

    @override
    def _table_drop(self, table, **kwargs):
        return self.execute(f"DROP TABLE {table}")

    @override
    def _table_desc(self, table, **kwargs):
        return self.query(f"DESCRIBE {table}", **kwargs)

    @override
    def _table_partition_cols(self, table, **kwargs):
        desc = self._table_desc(table, **kwargs)
        if "Extra" in desc:
            return list(desc[desc["Extra"].str.contains("partition key")]["Column"])
        return []

    @override
    def _table_head(self, table, n=10, **kwargs):
        return self.query(f"SELECT * FROM {table} LIMIT {n}", **kwargs)

    @override
    def _table_props(self, table, **kwargs):
        raise NotImplementedError



import pickle

from omniduct.caches._serializers import Serializer


class CursorSerializer(Serializer):
    """
    Serializes and deserializes cursor objects for use with the Cache.
    """

    @property
    def file_extension(self):
        """str: The file extension to use when storing in the cache."""
        return ".pickled_cursor"

    def serialize(self, obj, fh):
        """
        Serialize a cursor object into a nominated file handle.

        Args:
            obj (DB-API 2.0 cursor): The cursor to serialize.
            fh (binary file handle): A file-like object opened in binary mode
                capable of being written into.
        """
        description = obj.description
        rows = obj.fetchall()
        pickle.dump((description, rows), fh)

    def deserialize(self, fh):
        """
        Deserialize a cursor object into a DB-API 2.0 compatible cursor object.

        Args:
            fh (binary file handle): A file-like object from which serialized
                data can be read.

        Returns:
            CachedCursor: A CacheCursor object representing a previously
                serialized cursor.
        """
        description, rows = pickle.load(fh)
        return CachedCursor(description, rows)


class CachedCursor:
    """
    A DB-API 2.0 cursor implementation atop of static data.

    This class is used to present reconstituted data cached from a cursor object
    in a form compatible with the original cursor object.
    """

    def __init__(self, description, rows):
        self._description = description
        self._rows = rows

        self._iter = None

    @property
    def iter(self):
        if not getattr(self, "_iter"):
            self._iter = (row for row in self._rows)
        return self._iter

    arraysize = 1

    @property
    def description(self):
        return self._description

    @property
    def row_count(self):
        return -1

    def close(self):
        pass

    def execute(self, operation, parameters=None):
        raise NotImplementedError(
            "Cached cursors are not connected to a database, and cannot be "
            "used for database operations."
        )

    def executemany(self, operation, seq_of_parameters=None):
        raise NotImplementedError(
            "Cached cursors are not connected to a database, and cannot be "
            "used for database operations."
        )

    def fetchone(self):
        return next(self.iter)

    def fetchmany(self, size=None):
        size = size or self.arraysize
        return [self.fetchone() for _ in range(size)]

    def fetchall(self):
        return list(self.iter)

    def setinputsizes(self, sizes):
        pass

    def setoutputsize(self, size, column=None):
        pass



from __future__ import absolute_import

from interface_meta import override

from omniduct.utils.debug import logger

from .base import DatabaseClient


class DruidClient(DatabaseClient):
    """
    This Duct connects to a Druid server using the `pydruid` python library.
    """

    PROTOCOLS = ["druid"]
    DEFAULT_PORT = 80
    NAMESPACE_NAMES = ["table"]
    NAMESPACE_QUOTECHAR = '"'
    NAMESPACE_SEPARATOR = "."

    @override
    def _init(self):
        self.__druid = None

    # Connection
    @override
    def _connect(self):
        from pydruid.db import connect  # pylint: disable=import-error

        logger.info("Connecting to Druid database ...")
        self.__druid = connect(  # pylint: disable=attribute-defined-outside-init
            self.host, self.port, path="/druid/v2/sql/", scheme="http"
        )
        if self.username or self.password:
            logger.warning(
                "Duct username and password not passed to pydruid connection. "
                "pydruid connection currently does not allow these fields to be passed."
            )

    @override
    def _is_connected(self):
        return self.__druid is not None

    @override
    def _disconnect(self):
        logger.info("Disconnecting from Druid database ...")
        try:
            self.__druid.close()
        except:  # pylint: disable=bare-except
            pass
        self.__druid = None  # pylint: disable=attribute-defined-outside-init

    # Querying
    @override
    def _execute(self, statement, cursor, wait, session_properties):
        cursor = cursor or self.__druid.cursor()
        cursor.execute(statement)
        return cursor

    @override
    def _table_list(self, namespace, like=None, **kwargs):
        cmd = "SELECT * FROM INFORMATION_SCHEMA.TABLES"
        return self.query(cmd, **kwargs)

    @override
    def _table_exists(self, table, **kwargs):
        logger.disabled = True
        try:
            self.table_desc(table, **kwargs)
            return True
        except:  # pylint: disable=bare-except
            return False
        finally:
            logger.disabled = False

    @override
    def _table_drop(self, table, **kwargs):
        raise NotImplementedError

    @override
    def _table_desc(self, table, **kwargs):
        query = f"\n            SELECT\n                TABLE_SCHEMA\n                , TABLE_NAME\n                , COLUMN_NAME\n                , ORDINAL_POSITION\n                , COLUMN_DEFAULT\n                , IS_NULLABLE\n                , DATA_TYPE\n            FROM INFORMATION_SCHEMA.COLUMNS\n            WHERE TABLE_NAME = '{table}'"
        return self.query(query, **kwargs)

    @override
    def _table_head(self, table, n=10, **kwargs):
        return self.query(f"SELECT * FROM {table} LIMIT {n}", **kwargs)

    @override
    def _table_props(self, table, **kwargs):
        raise NotImplementedError



from __future__ import absolute_import

import urllib

from interface_meta import override

from omniduct.utils.debug import logger

from .base import DatabaseClient
from ._schemas import SchemasMixin
from . import _pandas


class SQLAlchemyClient(DatabaseClient, SchemasMixin):
    """
    This Duct connects to several different databases using one of several
    SQLAlchemy drivers. In general, these are provided for their potential
    utility, but will be less functional than the specially crafted database
    clients.
    """

    PROTOCOLS = [
        "sqlalchemy",
        "firebird",
        "mssql",
        "mysql",
        "oracle",
        "postgresql",
        "sybase",
        "snowflake",
    ]
    NAMESPACE_NAMES = ["database", "table"]
    NAMESPACE_QUOTECHAR = '"'  # TODO: Apply overrides depending on protocol?
    NAMESPACE_SEPARATOR = "."

    @property
    @override
    def NAMESPACE_DEFAULT(self):
        return {"database": self.database}

    @override
    def _init(self, dialect=None, driver=None, database="", engine_opts=None):
        assert self._port is not None, (
            "Omniduct requires SQLAlchemy databases to manually specify a port, as "
            "it will often be the case that ports are being forwarded."
        )

        if self.protocol != "sqlalchemy":
            self.dialect = self.protocol
        else:
            self.dialect = dialect
        assert self.dialect is not None, "Dialect not specified."

        self.driver = driver
        self.database = database
        self.connection_fields += ("schema",)
        self.engine_opts = engine_opts or {}

        self.engine = None
        self.connection = None

    @property
    def db_uri(self):
        # pylint: disable-next=consider-using-f-string
        return "{dialect}://{login}@{host_port}/{database}".format(
            dialect=self.dialect + (f"+{self.driver}" if self.driver else ""),
            login=self.username
            + (f":{urllib.parse.quote_plus(self.password)}" if self.password else ""),
            host_port=self.host + (f":{self.port}" if self.port else ""),
            database=self.database,
        )

    @property
    def _sqlalchemy_engine(self):
        """
        The SQLAlchemy engine object for the SchemasMixin.
        """
        return self.engine
    
    @_sqlalchemy_engine.setter
    def _sqlalchemy_engine(self, engine):
        self.engine = engine

    @override
    def _connect(self):
        import sqlalchemy

        if self.protocol not in ["mysql"]:
            logger.warning(
                "While querying and executing should work as "
                "expected, some operations on this database client "
                "(such as listing tables, querying to tables, etc) "
                "may not function as expected due to the backend "
                "not supporting ANSI SQL."
            )

        # pylint: disable-next=attribute-defined-outside-init
        self.engine = sqlalchemy.create_engine(self.db_uri, **self.engine_opts)

    @override
    def _is_connected(self):
        return self.engine is not None

    @override
    def _disconnect(self):
        # pylint: disable-next=attribute-defined-outside-init
        self.engine = None
        # pylint: disable-next=attribute-defined-outside-init
        self._schemas = None

    @override
    def _execute(
        self, statement, cursor, wait, session_properties, query=True, **kwargs
    ):
        assert wait, "`SQLAlchemyClient` does not support asynchronous operations."
        if cursor:
            cursor.execute(statement)
        else:
            cursor = self.engine.execute(statement).cursor
        return cursor

    @override
    def _query_to_table(self, statement, table, if_exists, **kwargs):
        statements = []

        if if_exists == "fail" and self.table_exists(table):
            raise RuntimeError(f"Table {table} already exists!")
        if if_exists == "replace":
            statements.append(f"DROP TABLE IF EXISTS {table};")
        elif if_exists == "append":
            raise NotImplementedError(
                f"Append operations have not been implemented for {self.__class__.__name__}."
            )

        statement = f"CREATE TABLE {table} AS ({statement})"
        return self.execute(statement, **kwargs)

    @override
    def _dataframe_to_table(self, df, table, if_exists="fail", **kwargs):
        return _pandas.to_sql(
            df=df,
            name=table.table,
            schema=table.database,
            con=self.engine,
            index=False,
            if_exists=if_exists,
            **kwargs,
        )

    @override
    def _table_list(self, namespace, **kwargs):
        return self.query(f"SHOW TABLES IN {namespace}", **kwargs)

    @override
    def _table_exists(self, table, **kwargs):
        logger.disabled = True
        try:
            self.table_desc(table, **kwargs)
            return True
        except:  # pylint: disable=bare-except
            return False
        finally:
            logger.disabled = False

    @override
    def _table_drop(self, table, **kwargs):
        return self.execute(f"DROP TABLE {table}")

    @override
    def _table_desc(self, table, **kwargs):
        return self.query(f"DESCRIBE {table}", **kwargs)

    @override
    def _table_head(self, table, n=10, **kwargs):
        return self.query(f"SELECT * FROM {table} LIMIT {n}", **kwargs)

    @override
    def _table_props(self, table, **kwargs):
        raise NotImplementedError



from abc import ABCMeta, abstractmethod

from interface_meta import inherit_docs


def process_line_arguments(f):
    def wrapped(*args, **kwargs):
        args = list(args)
        new_args, new_kwargs = _process_line_arguments(args.pop(0))
        args += new_args
        kwargs.update(new_kwargs)
        return f(*args, **kwargs)

    return wrapped


def process_line_cell_arguments(f):
    def wrapped(*args, **kwargs):
        args = list(args)
        new_args, new_kwargs = _process_line_arguments(args.pop(0))
        if len(args) == 0:
            args += [None]
        args += new_args
        kwargs.update(new_kwargs)
        return f(*args, **kwargs)

    return wrapped


def _process_line_arguments(line_arguments):
    from IPython import get_ipython

    args = []
    kwargs = {}
    reached_kwargs = False
    for arg in line_arguments.split():
        if "=" in arg:
            reached_kwargs = True
            key, value = arg.split("=")
            value = eval(value, get_ipython().user_ns)
            if key in kwargs:
                raise ValueError(f"Duplicate keyword argument `{key}`.")
            kwargs[key] = value
        else:
            if reached_kwargs:
                raise ValueError(f"Positional argument `{arg}` after keyword argument.")
            args.append(arg)
    return args, kwargs


class MagicsProvider(metaclass=ABCMeta):
    @inherit_docs("_register_magics")
    def register_magics(self, base_name=None):
        base_name = base_name or self.name
        if base_name is None:
            raise RuntimeError("Cannot register magics without a base_name.")

        try:
            from IPython import get_ipython

            ip = get_ipython()
            assert ip is not None
            has_ipython = True
        except Exception:  # pylint: disable=broad-exception-caught
            has_ipython = False

        if has_ipython:
            self._register_magics(base_name)

    @abstractmethod
    def _register_magics(self, base_name):
        pass



import pytest

from omniduct.databases._namespaces import ParsedNamespaces


class TestParseNamespaces:
    def test_simple(self):
        namespace = ParsedNamespaces.from_name(
            name="my_db.my_table", namespaces=["database", "table"]
        )

        assert namespace.database == "my_db"
        assert namespace.table == "my_table"
        assert namespace.as_dict() == {"database": "my_db", "table": "my_table"}

    def test_quoted_names(self):
        namespace = ParsedNamespaces.from_name(
            name="`my_db`.`my . table`",
            namespaces=["catalog", "database", "table"],
            quote_char="`",
        )

        assert namespace.catalog is None
        assert namespace.database == "my_db"
        assert namespace.table == "my . table"
        assert namespace.as_dict() == {
            "catalog": None,
            "database": "my_db",
            "table": "my . table",
        }

    def test_separator(self):
        namespace = ParsedNamespaces.from_name(
            name="cat|my_db|my_table",
            namespaces=["catalog", "database", "table"],
            separator="|",
        )

        assert namespace.catalog == "cat"
        assert namespace.database == "my_db"
        assert namespace.table == "my_table"
        assert namespace.as_dict() == {
            "catalog": "cat",
            "database": "my_db",
            "table": "my_table",
        }

    def test_parsing_failure(self):
        with pytest.raises(ValueError):
            ParsedNamespaces.from_name(name="my_db.my_table", namespaces=["table"])

    def test_nonexistent_namespace(self):
        with pytest.raises(AttributeError):
            ParsedNamespaces.from_name(name="my_table", namespaces=["table"]).database

    def test_not_encapsulated(self):
        namespace = ParsedNamespaces.from_name("my_db.my_table", ["database", "table"])
        assert namespace.as_dict() == {"database": "my_db", "table": "my_table"}

        with pytest.raises(ValueError):
            ParsedNamespaces.from_name(namespace, ["schema", "table"])

    def test_empty(self):
        namespace = ParsedNamespaces.from_name("", ["database", "table"])

        assert bool(namespace) is False
        assert namespace.database is None
        assert namespace.table is None
        assert str(namespace) == ""
        assert repr(namespace) == "Namespace<>"

    def test_parent(self):
        namespace = ParsedNamespaces.from_name(
            name="my_db.my_table", namespaces=["catalog", "database", "table"]
        )

        assert namespace.parent.name == '"my_db"'
        assert namespace.parent.as_dict() == {"catalog": None, "database": "my_db"}

    def test_casting(self):
        namespace = ParsedNamespaces.from_name(
            name="my_db.my_table", namespaces=["catalog", "database", "table"]
        )

        assert str(namespace) == '"my_db"."my_table"'
        assert bool(namespace) is True
        assert (
            namespace.__bool__() == namespace.__nonzero__()
        )  # Python 2/3 compatibility
        assert repr(namespace) == 'Namespace<"my_db"."my_table">'
